using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography.X509Certificates;
using Jypeli;
using Jypeli.Assets;
using Jypeli.Controls;
using Jypeli.Widgets;

namespace matopeli;

/// @author Paulus Kopperoinen
/// @version 30.10.2023
/// <summary>
/// Perinteinen matopeli, jossa tavoitteena kasvattaa matoa mahdollisimman suureksi omenia syömällä.
/// </summary>

public class matopeli : PhysicsGame
{
    private static readonly string[] tasonkuva = {
        "#####################",
        "#                   #",
        "#                   #",
        "#                   #",
        "#                   #",
        "#                   #",
        "#         m         #",
        "#                   #",
        "#                   #",
        "#                   #",
        "#                   #", 
        "#                   #",
        "#                   #",
        "#                   #",
        "#                   #",
        "#                   #",
        "#o                  #",
        "#####################",
    };
    private Objektit Mato;
    private Objektit Omena;
    private Objektit MatoMuu;
    
    public IntMeter pistelaskuri;
    public IntMeter Ennatyslaskuri;
    
    public List<Vector> MatoSijainti = new List<Vector>();
    public List<PhysicsObject> MatoKeho = new List<PhysicsObject>();

    public int leveys;
    public int korkeus;
    public double Nopeus;

    private Vector OmenaAlkuPaikka;
    
    private static readonly Image seinakuva = LoadImage("seina.png");
    private static readonly Image matopaakuva = LoadImage("matopaa.png");
    private readonly Image matokehokuva = LoadImage("matokeho.png");
    private readonly Image omenakuva = LoadImage("omena.png");
    
    private bool SuuntaYlos;
    private bool SuuntaAlas;
    private bool SuuntaOikea;
    private bool SuuntaVasen;

    private bool MatoLiikkunut;

    private bool OmenaKentalla;
    
    
    public override void Begin()
    {
        AloitaUusiPeli();
    }

    
    /// <summary>
    /// Poistaa vanhan pelin tiedot ja alustaa uuden pelin
    /// </summary>
    public void AloitaUusiPeli()
    {
        MatoKeho.Clear();
        MatoSijainti.Clear();
        ClearAll();
        Asetukset();
        TileMap tiles = TileMap.FromStringArray(tasonkuva);
        tiles['#'] = LuoSeina;
        tiles['m'] = LuoMato;
        tiles.Insert(leveys,korkeus); ;
        OmenaAlkuPaikka = Mato.Position;
        Asetaohjaimet();
        LuoAikalaskuri();
        LuoPistelaskuri();
        LuoOmena();
    }
    
    
    /// <summary>
    /// Määrittää ohjaimet, joita pelaaja voi käyttää ja mitä niitä painamalla tapahtuu
    /// </summary>
    private void Asetaohjaimet()
    {
        Keyboard.Listen(Key.Up, ButtonState.Pressed, VaihdaSuuntaYlos, "Turn upwards");
        Keyboard.Listen(Key.Down, ButtonState.Pressed, VaihdaSuuntaAlas, "Turn downwards");
        Keyboard.Listen(Key.Left, ButtonState.Pressed, VaihdaSuuntaVasen , "Turn left");
        Keyboard.Listen(Key.Right, ButtonState.Pressed, VaihdaSuuntaOikea, "Turn right");
        Keyboard.Listen(Key.Enter, ButtonState.Pressed, LiikutaMatoa, "New Game");
    }


    /// <summary>
    /// Laskee pisteet, joita kerätään omenoita syömällä
    /// </summary>
    void LuoPistelaskuri()
    {
        pistelaskuri = new IntMeter(0);               
        
        Label pistenaytto = new Label(); 
        pistenaytto.X = Screen.Left + 500;
        pistenaytto.Y = Screen.Top - 10;
        pistenaytto.TextColor = Color.Black;
        pistenaytto.Color = Color.White;
        pistenaytto.Title = "Pisteet: ";

        pistenaytto.BindTo(pistelaskuri);
        Add(pistenaytto);
    }
    

    /// <summary>
    /// Pelin asetukset
    /// </summary>
    private void Asetukset()
    {
    leveys = 40;
    korkeus = 40;
    SuuntaYlos = false;
    SuuntaAlas = false;
    SuuntaOikea = true; 
    SuuntaVasen = false;
    Nopeus = 0.1;
    MatoLiikkunut = false;
    OmenaKentalla = true;
    }
    
    
    /// <summary>
    /// Päättää pelin ja aloitaa uuden, kun mato kuolee
    /// </summary>
    private void MatoKuolee()
    {
        AloitaUusiPeli();
    }
    
    /// <summary>
    /// Luo madon kentälle
    /// </summary>
    /// <returns></returns>
    private PhysicsObject LuoMato()
    {
        Mato = new Objektit(leveys-5, korkeus-5);
        Mato.Image = matopaakuva;
        Add(Mato);
        Mato.Tag = "mato";
        
        AddCollisionHandler(Mato, "omena", delegate(IPhysicsObject Mato, IPhysicsObject Omena) {MatoSyoOmenan();  });
        AddCollisionHandler(Mato, "seina", delegate(IPhysicsObject Mato, IPhysicsObject seina) {MatoKuolee();  });
        AddCollisionHandler(Mato, "matokeho", delegate(IPhysicsObject Mato, IPhysicsObject matokeho) {MatoKuolee();  });
        
        return Mato;
    }

    
    /// <summary>
    /// Luodaan omena johonkin satunnaiseen kohtaan kartalla, joka ei ole samassa kohtaa madon kanssa.
    /// </summary>
    /// <returns></returns>
    private PhysicsObject LuoOmena()
    {
        Vector paikka = ArvoOmenaPaikka();
        Omena = new Objektit(korkeus-10, leveys-10);
        Omena.Shape = Shape.Circle;
        Omena.Position = paikka;
        Omena.Image = omenakuva;
        Omena.Tag = "omena";
        Add(Omena);
        OmenaKentalla = true;
        return Omena;
    }

    
    /// <summary>
    /// tehtävä on arpoa omenalle paikka, johon se luodaan. En ole vieläkään varma kuinka hyvin tämä toimii.
    /// </summary>
    /// <returns></returns>
    private Vector ArvoOmenaPaikka()
    {
        Vector OmenaPaikka;
        double x = RandomGen.NextInt(-8, 8);
        x = OmenaAlkuPaikka.X + x * leveys;
        double y = RandomGen.NextInt(-10, 5);
        y = OmenaAlkuPaikka.Y + y * leveys;
        
        x = RandomGen.NextInt(-8, 8);
        x = OmenaAlkuPaikka.X + x * leveys;
            
        y = RandomGen.NextInt(-10, 5);
        y = OmenaAlkuPaikka.Y + y * leveys;

        for (int i = 0; i < 100; i++)
        {
            y = RandomGen.NextInt(-10, 5);
            y = OmenaAlkuPaikka.Y + y * leveys;
            
            if (x-Mato.X<leveys)
            {
                if (y-Mato.Y<leveys)
                {
                    if (MatoKeho.Count>0)
                    {
                        if (x-MatoMuu.X<leveys)
                        {
                            if (y-MatoMuu.Y<leveys)
                            {
                                break;
                            }
                        }
                    }
                    else break;
                }
            }
        }
        OmenaPaikka = new Vector(x, y);
        return OmenaPaikka;
    }
        
    
    /// <summary>
    /// Luo kentän reunat
    /// </summary>
    /// <returns></returns>
    private PhysicsObject LuoSeina()
    {
        PhysicsObject Seina = new PhysicsObject(leveys, korkeus);
        Seina.MakeStatic();
        Seina.Color = Color.Green;
        Seina.Image = seinakuva;
        Seina.Tag = "seina";
        return Seina;
    }
    
    
    /// <summary>
    /// Tehdään aliohjelma, joka luo uusia osia madon perään.
    /// </summary>
    public PhysicsObject LuoMatoKeho()
    {
        MatoMuu = new Objektit(leveys-6, korkeus-6);
        MatoMuu.Tag = "matokeho";
        MatoMuu.Image = matokehokuva;
        Vector Paikka = MatoSijainti[MatoSijainti.Count -1 - MatoKeho.Count];
        MatoMuu.Position = Paikka;
        Add(MatoMuu);
        MatoKeho.Add(MatoMuu);
        return MatoMuu;
    }
    
    
    /// <summary>
    /// Tekee tarvittavat toimet, kun mato onnistuu syömään omenan
    /// </summary>
    private void MatoSyoOmenan()
    {
        Omena.Destroy();
        pistelaskuri.Value += 1;
        OmenaKentalla = false;
    }
    
    
    /// <summary>
    /// Antaa tiedon LiikutaMatoa aliohjelmalle, että madon pitäisi kääntyä ylös.
    /// </summary>
     private void VaihdaSuuntaYlos()
        {
            
            if (SuuntaAlas==true)
            {
                SuuntaOikea = false;
                SuuntaVasen = false;
                SuuntaYlos = false;
                SuuntaAlas = true;
            }
            else
            {
                SuuntaOikea = false;
                SuuntaVasen = false;
                SuuntaYlos = true;
                SuuntaAlas = false;
            }
        }
     
    
    /// <summary>
    /// Antaa tiedon LiikutaMatoa aliohjelmalle, että madon pitäisi kääntyä alas.
    /// </summary>
     private void VaihdaSuuntaAlas()
        {
            if (SuuntaYlos==true)
            {
                SuuntaOikea = false;
                SuuntaVasen = false;
                SuuntaYlos = true;
                SuuntaAlas = false;
            }
            else
            {
                SuuntaOikea = false;
                SuuntaVasen = false;
                SuuntaYlos = false;
                SuuntaAlas = true;
            }
        }
      
     
    /// <summary>
    /// Antaa tiedon LiikutaMatoa aliohjelmalle, että madon pitäisi kääntyä oikeaan.
    /// </summary>
     private void VaihdaSuuntaOikea()
        {
            if (SuuntaVasen == true)
            {
                
                SuuntaOikea = false;
                SuuntaVasen = true;
                SuuntaYlos = false;
                SuuntaAlas = false;
            }
            else
            {
                SuuntaOikea = true;
                SuuntaVasen = false;
                SuuntaYlos = false;
                SuuntaAlas = false;
            }

        }
       
     
    /// <summary>
    /// Antaa tiedon LiikutaMatoa aliohjelmalle, että madon pitäisi kääntyä vasempaan.
    /// </summary>
     private void VaihdaSuuntaVasen()
        {
            if (SuuntaOikea==true)
            {
                SuuntaOikea = true;
                SuuntaVasen = false;
                SuuntaYlos = false;
                SuuntaAlas = false;
            }
            else
            {
                SuuntaOikea = false;
                SuuntaVasen = true;
                SuuntaYlos = false;
                SuuntaAlas = false;
            }
        }
     
     
     /// <summary>
     /// Aikalaskuri toimii pelin moottorina ja antaa PaivitaKentälle tiedon, kun pitää tarkistaa mitä kentällä on tapahtunut.
     /// </summary>
     private void LuoAikalaskuri()
        {
            Timer aikalaskuri = new Timer();
            aikalaskuri.Interval=Nopeus;
            aikalaskuri.Start();
            aikalaskuri.Timeout += PaivitaKentta; 
        }

     
     /// <summary>
     /// Kun ajastimesta loppuu aika, siirtyy peli tähän funktioon, jonka tarkoituksena on hahmottaa, mitä kaikkea pelissä on tapahtunut ja miten siihen tulee reagoida.
     /// </summary>
     private void PaivitaKentta()
     {
         if (MatoLiikkunut == false)
         {
             return;
         }
         
         if (OmenaKentalla == false)
         {
             LuoMatoKeho();
             LuoOmena();
         }
         
         MatoSijainti.Add(new Vector(Mato.X,Mato.Y));
         LiikutaMatoa();
         LiikutaKehoa();
         MatoSijainti.Add(new Vector(Mato.X,Mato.Y));  
     }
     
     
     /// <summary>
     /// LiikutaMatoa aliohjelman tehtävä on selvittää tieto siitä, mihin suuntaan on tarkoitus liikkua ja liikuttaa matoa siihen suuntaan.
     /// </summary>
    private void LiikutaMatoa()
    {
        MatoLiikkunut = false;
        if (SuuntaYlos==true)
        {
            Mato.Y = Mato.Y + leveys;
            Mato.Angle = Angle.FromDegrees(90);
        }
        if (SuuntaAlas==true)
        {
            Mato.Y= Mato.Y - leveys;
            Mato.Angle = Angle.FromDegrees(270);
        }
        if (SuuntaOikea==true)
        {
            Mato.X= Mato.X + leveys;
            Mato.Angle = Angle.FromDegrees(0);
           
        }
        if (SuuntaVasen==true)
        {
            Mato.X= Mato.X - leveys;
            Mato.Angle = Angle.FromDegrees(180);
            
        }

        MatoLiikkunut = true;
    }

    
    /// <summary>
    /// Liikutakehoa, aliohjelman tehtävänä on liikuttaa madon kehoa sen pään mukaisesti.
    /// </summary>
    private void LiikutaKehoa()
    {
        if (MatoKeho.Count==0)
        {
            return;
        }
        MatoLiikkunut = false;
        MatoKeho[0].X = MatoSijainti[MatoSijainti.Count - 1].X;
        MatoKeho[0].Y = MatoSijainti[MatoSijainti.Count - 1].Y;
        
        for (int i = 1; i < MatoKeho.Count; i++)
        {
            MatoKeho[i].X = MatoSijainti[MatoSijainti.Count - 2*i].X;
            MatoKeho[i].Y = MatoSijainti[MatoSijainti.Count - 2*i].Y;
        }

        MatoLiikkunut = true;
    }


    public class Objektit : PhysicsObject
    {
         public Objektit(double leveys, double korkeus)
        : base(leveys, korkeus)
        {
            Mass = 1000;
            IsUpdated = true;
            IgnoresCollisionResponse = true;
        }
    }
}

        





        




    